# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from sqlalchemy import MetaData, String, Table


def upgrade(migrate_engine):
    meta = MetaData()
    meta.bind = migrate_engine

    for table_name in ['profile', 'policy', 'cluster', 'node', 'receiver',
                       'credential', 'action', 'event']:
        table = Table(table_name, meta, autoload=True)
        table.c.user.alter(type=String(64))
        table.c.project.alter(type=String(64))
        if table_name not in ['credential', 'event']:
            table.c.domain.alter(type=String(64))
